# coding: utf-8

"""
    Kiln API Specifications

    This API provides reporting staking data on various protocols as well as network wide data, staking transaction crafting features and so on. In order to use it, you should first get an API token from your Kiln dashboard (applications section). If you don't have access to our dashboard, please get in touch at hello@kiln.fi. Once you have your API token, you can set it as a bearer token in your request headers.  ### Backward Compatibility  This is an experimental specification used for development and testing, do not rely on what is here unless you know the implications.  The official Kiln API specification following backward compatible changes can be found [here](/).  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Contact: contact@kiln.fi
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest

import kiln_connect.openapi_client
from kiln_connect.openapi_client.api.eth_api import EthApi  # noqa: E501
from kiln_connect.openapi_client.rest import ApiException


class TestEthApi(unittest.TestCase):
    """EthApi unit test stubs"""

    def setUp(self):
        self.api = kiln_connect.openapi_client.api.eth_api.EthApi()  # noqa: E501

    def tearDown(self):
        pass

    def test_get_eth_kiln_stats(self):
        """Test case for get_eth_kiln_stats

        Kiln Stats  # noqa: E501
        """
        pass

    def test_get_eth_network_stats(self):
        """Test case for get_eth_network_stats

        Network Stats  # noqa: E501
        """
        pass

    def test_get_eth_operations(self):
        """Test case for get_eth_operations

        Operations  # noqa: E501
        """
        pass

    def test_get_eth_reports(self):
        """Test case for get_eth_reports

        Excel Reports  # noqa: E501
        """
        pass

    def test_get_eth_rewards(self):
        """Test case for get_eth_rewards

        Rewards  # noqa: E501
        """
        pass

    def test_get_eth_stakes(self):
        """Test case for get_eth_stakes

        Stakes  # noqa: E501
        """
        pass

    def test_get_eth_tx_status(self):
        """Test case for get_eth_tx_status

        Transaction Status  # noqa: E501
        """
        pass

    def test_post_eth_broadcast_tx(self):
        """Test case for post_eth_broadcast_tx

        Broadcast Transaction  # noqa: E501
        """
        pass

    def test_post_eth_keys(self):
        """Test case for post_eth_keys

        Validation Keys  # noqa: E501
        """
        pass

    def test_post_eth_prepare_tx(self):
        """Test case for post_eth_prepare_tx

        Prepare Transaction  # noqa: E501
        """
        pass

    def test_post_eth_stake_tx(self):
        """Test case for post_eth_stake_tx

        Stake Transaction  # noqa: E501
        """
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    Kiln API Specifications

    This API provides reporting staking data on various protocols as well as network wide data, staking transaction crafting features and so on. In order to use it, you should first get an API token from your Kiln dashboard (applications section). If you don't have access to our dashboard, please get in touch at hello@kiln.fi. Once you have your API token, you can set it as a bearer token in your request headers.  ### Backward Compatibility  This is an experimental specification used for development and testing, do not rely on what is here unless you know the implications.  The official Kiln API specification following backward compatible changes can be found [here](/).  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Contact: contact@kiln.fi
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel
from kiln_connect.openapi_client.models.ethereum_broadcast_tx_response import EthereumBroadcastTxResponse

class PostEthBroadcastTx201Response(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    data: Optional[EthereumBroadcastTxResponse] = None
    __properties = ["data"]

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> PostEthBroadcastTx201Response:
        """Create an instance of PostEthBroadcastTx201Response from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of data
        if self.data:
            _dict['data'] = self.data.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> PostEthBroadcastTx201Response:
        """Create an instance of PostEthBroadcastTx201Response from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return PostEthBroadcastTx201Response.parse_obj(obj)

        _obj = PostEthBroadcastTx201Response.parse_obj({
            "data": EthereumBroadcastTxResponse.from_dict(obj.get("data")) if obj.get("data") is not None else None
        })
        return _obj




"""Ethereum commands.

This file contains multiple CLI commands showcasing how to use the
KilnConnect SDK to interact with the Ethereum blockchain.

Code here is voluntarily kept simple: it could be refactored with some
levels of abstraction to avoid repetitions, but would imply readers to
understand things unrelated to what their primary goal is: use the
SDK. So let's keep it stupid simple so the integration work is
simpler.
"""

import click
import os
import re
import rich
import typer
import time

import fireblocks_sdk
from rich.console import Console
from rich.table import Table
from typing import Optional, Tuple

import kiln_connect


eth_cli = typer.Typer(
    name='eth', help='Staking utilities for Ethereum', no_args_is_help=True)


console = Console()
error_console = Console(stderr=True)


def pretty_wei_to_eth(wei: str) -> str:
    """Quick helper to pretty print WEI to ETH.
    """
    if not wei:
        return 'n/a'
    eth = str(round(int(wei) / 1e18, 3))
    return f"{eth}Ξ"


def sort_identifiers(identifiers: list[str]) -> Tuple[list[str], list[str], list[str]]:
    """Sorts ETH filtering identifiers in corresponding buckets.

    The Kiln ETH API supports filtering by:

    - ETH2 BLS validator address (consensus address)
    - ETH1 wallet address (execution address)
    - Kiln Account ID (UUID)

    This functions returns the identifiers sorted in their corresponding bucket.
    """

    validators = []
    accounts = []
    wallets = []

    for identifier in identifiers:
        identifier = identifier[2:] if '0x' in identifier else identifier

        # Validator addresses are 96 bytes
        if re.match('^[0-9a-zA-Z]{96}$', identifier):
            validators.append(identifier)
            continue

        # Kiln Account UUIDs
        if re.match('^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$', identifier):
            accounts.append(identifier)
            continue

        # Wallet addresses are 40 bytes
        if re.match('^[0-9a-zA-Z]{40}$', identifier):
            wallets.append(identifier)
            continue

        raise click.UsageError(
            "Unknown identifier (should be a BLS validator address, an ETH wallet address or a Kiln account UUID)")

    # This is a current limit of the Kiln API, we only support one
    # filter type at a time. This removes confusion when a staked is
    # matched by multiple filters.
    if (validators and accounts) or (validators and wallets) or (accounts and wallets):
        raise click.UsageError(
            "Identifiers should be of the same type"
        )

    return validators or None, accounts or None, wallets or None


@eth_cli.command("stakes")
def ethereum_stakes(identifiers: list[str]):
    """List Ethereum Stake status.
    """
    with kiln_connect.KilnConnect(kiln_connect.KilnConfig.from_env()) as kc:
        validators, accounts, wallets = sort_identifiers(identifiers)

        stakes = kc.eth.get_eth_stakes(
            validators=validators, accounts=accounts, wallets=wallets).data

        table = Table('Stake(s)', 'Status', 'Balance', 'Rewards')
        for stake in stakes:
            table.add_row(
                stake.validator_address,
                stake.state,
                pretty_wei_to_eth(stake.balance),
                pretty_wei_to_eth(stake.rewards))

        console.print(table)


@eth_cli.command("rewards")
def ethereum_rewards(identifiers: list[str]):
    """View Ethereum rewards.
    """
    with kiln_connect.KilnConnect(kiln_connect.KilnConfig.from_env()) as kc:
        validators, accounts, wallets = sort_identifiers(identifiers)
        rewards = kc.eth.get_eth_rewards(
            validators=validators, accounts=accounts, wallets=wallets).data

        table = Table(
            'Time', 'Stake Balance', 'Consensus', 'Execution', 'Rewards', 'Gross APY')
        for reward in rewards:
            table.add_row(
                str(reward.var_date),
                pretty_wei_to_eth(reward.stake_balance),
                pretty_wei_to_eth(reward.consensus_rewards),
                pretty_wei_to_eth(reward.execution_rewards),
                pretty_wei_to_eth(reward.rewards),
                str(round(reward.gross_apy, 3)))

        console.print(table)


@eth_cli.command("operations")
def ethereum_operations(identifiers: list[str]):
    """List Ethereum operations.
    """
    with kiln_connect.KilnConnect(kiln_connect.KilnConfig.from_env()) as kc:
        validators, accounts, wallets = sort_identifiers(identifiers)
        operations = kc.eth.get_eth_operations(
            validators=validators, accounts=accounts, wallets=wallets).data

        table = Table('Stake', 'Time', 'Type')
        for op in operations:
            op = op.actual_instance
            table.add_row(
                op.validator_address,
                str(op.time),
                op.type)

        console.print(table)


@eth_cli.command("network-stats")
def ethereum_network_stats():
    """Show Ethereum Network Stats.
    """
    with kiln_connect.KilnConnect(kiln_connect.KilnConfig.from_env()) as kc:
        ns = kc.eth.get_eth_network_stats()

        table = Table('Network Gross APY %', 'Supply Staked %')
        table.add_row(
            str(round(ns.data.network_gross_apy, 3)),
            str(round(ns.data.supply_staked_percent, 3)))

        console.print(table)


@eth_cli.command("stake-via-fireblocks")
def ethereum_stake_via_fireblocks(account_id: str, wallet: str):
    """Stake ETH via fireblocks.
    """
    with kiln_connect.KilnConnect(kiln_connect.KilnConfig.from_env()) as kc:
        kc.eth.stake('fireblocks', account_id, wallet, 32000000000000000000)



"""Accounts wrapper class.

This Accounts class is meant to accessed via KilnConnect.accounts. It
contains helpers around accounts management.
"""

from kiln_connect.integrations import Integrations
from kiln_connect.openapi_client import (
    ApiClient,
)
from kiln_connect.openapi_client import (
    AccountsApi,
)


class Accounts(AccountsApi):
    """Wrapper for the Accounts API.
    """

    def __init__(self, api: ApiClient, integrations: Integrations):
        super().__init__(api)
        self.integrations = integrations



"""Wrapper a bit similar to the Typescript SDK with signing integrations.
"""

import time
from dataclasses import dataclass

import fireblocks_sdk

from .errors import *


@dataclass
class IntegrationConfig:
    """Configuration of a Kiln integration.
    """
    name: str
    provider: str
    parameters: dict


class FireblocksIntegration:
    """Handles the Fireblocks integration.
    """

    def __init__(self, config: IntegrationConfig):
        self.fireblocks = None

        self.src = fireblocks_sdk.TransferPeerPath(
            "VAULT_ACCOUNT", config.parameters['vault_account_id'])
        self.assets = config.parameters['assets']

        with open(config.parameters['raw_key_path'], 'r') as pk:
            d = pk.read()
            self.fb = fireblocks_sdk.FireblocksSDK(
                d, config.parameters['api_token'])

    def sign(self, asset: str, unsigned_tx_hash: str) -> dict:
        """Asks fireblocks to sign TX hash and wait for completion.
        """
        msg = fireblocks_sdk.RawMessage(
            [fireblocks_sdk.UnsignedMessage(unsigned_tx_hash)])
        asset = self.assets[asset]
        note = "Staked from Kiln SDK Python"

        sign_tx = self.fb.create_raw_transaction(msg, self.src, asset, note)
        tx_id = sign_tx.get("id")

        failed_states = [
            fireblocks_sdk.TRANSACTION_STATUS_BLOCKED,
            fireblocks_sdk.TRANSACTION_STATUS_FAILED,
            fireblocks_sdk.TRANSACTION_STATUS_REJECTED,
            fireblocks_sdk.TRANSACTION_STATUS_CANCELLED
        ]

        signed_tx = None
        while True:
            signed_tx = self.fb.get_transaction_by_id(tx_id)
            status = signed_tx.get('status')
            if status == fireblocks_sdk.TRANSACTION_STATUS_COMPLETED:
                break
            if status in failed_states:
                break
            time.sleep(1)

        signatures = signed_tx.get("signedMessages")
        if not signatures or len(signatures) != 1:
            raise KilnFireblocksFailedToSignError()

        return signatures[0].get('signature')


class Integrations:
    """Handles integrations.

    For now it's kind of ad-hoc and we don't have a lot of
    integrations so we can keep it simple, once we start having a lot
    we can split each integration accordingly without changing this
    interface.
    """

    def __init__(self, configs: list[IntegrationConfig]):
        self.integrations = dict()

        for config in configs:
            if config.provider == 'fireblocks':
                i = FireblocksIntegration(config)
                self.integrations[config.name] = i

    def get_integration(self, name: str):
        """Returns the configured integration
        """
        return self.integrations[name]



# coding: utf-8

"""
    Kiln API Specifications

    This API provides reporting staking data on various protocols as well as network wide data, staking transaction crafting features and so on. In order to use it, you should first get an API token from your Kiln dashboard (applications section). If you don't have access to our dashboard, please get in touch at hello@kiln.fi. Once you have your API token, you can set it as a bearer token in your request headers.  ### Backward Compatibility  This is an experimental specification used for development and testing, do not rely on what is here unless you know the implications.  The official Kiln API specification following backward compatible changes can be found [here](/).  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Contact: contact@kiln.fi
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Optional
from pydantic import BaseModel, Field, StrictInt, StrictStr

class EthereumStake(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    validator_address: Optional[StrictStr] = Field(None, description="Public key of the validator")
    state: Optional[StrictStr] = Field(None, description="State of the Ethereum stake as seen be the consensus layer")
    activated_at: Optional[datetime] = Field(None, description="Date of activation on the Ethereum consensus layer")
    activated_epoch: Optional[StrictInt] = Field(None, description="Epoch of activation on the Ethereum consensus layer")
    delegated_block: Optional[StrictInt] = Field(None, description="Block at which the corresponding staking transaction was executed")
    deposit_tx_sender: Optional[StrictStr] = Field(None, description="Address of the sender of the first deposit transaction")
    execution_fee_recipient: Optional[StrictStr] = Field(None, description="Address of the last recipient of an execution reward")
    withdrawal_credentials: Optional[StrictStr] = Field(None, description="Ethereum withdrawal credentials")
    effective_balance: Optional[StrictStr] = Field(None, description="Effective balance in WEI of the stake as seen by the Ethereum consensus layer")
    balance: Optional[StrictStr] = Field(None, description="Current balance in WEI on the Ethereum consensus layer")
    consensus_rewards: Optional[StrictStr] = Field(None, description="Sum of consensus rewards in WEI earned by this stake")
    execution_rewards: Optional[StrictStr] = Field(None, description="Sum of execution rewards in WEI earned by this stake")
    rewards: Optional[StrictStr] = Field(None, description="Sum of consensus and execution rewards in WEI earned by this stake")
    gross_apy: Optional[float] = Field(None, description="Gross annual percentage yield")
    updated_at: Optional[datetime] = Field(None, description="Last date this data was updated")
    __properties = ["validator_address", "state", "activated_at", "activated_epoch", "delegated_block", "deposit_tx_sender", "execution_fee_recipient", "withdrawal_credentials", "effective_balance", "balance", "consensus_rewards", "execution_rewards", "rewards", "gross_apy", "updated_at"]

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> EthereumStake:
        """Create an instance of EthereumStake from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # set to None if activated_at (nullable) is None
        if self.activated_at is None:
            _dict['activated_at'] = None

        # set to None if activated_epoch (nullable) is None
        if self.activated_epoch is None:
            _dict['activated_epoch'] = None

        # set to None if delegated_block (nullable) is None
        if self.delegated_block is None:
            _dict['delegated_block'] = None

        # set to None if deposit_tx_sender (nullable) is None
        if self.deposit_tx_sender is None:
            _dict['deposit_tx_sender'] = None

        # set to None if execution_fee_recipient (nullable) is None
        if self.execution_fee_recipient is None:
            _dict['execution_fee_recipient'] = None

        # set to None if withdrawal_credentials (nullable) is None
        if self.withdrawal_credentials is None:
            _dict['withdrawal_credentials'] = None

        # set to None if effective_balance (nullable) is None
        if self.effective_balance is None:
            _dict['effective_balance'] = None

        # set to None if balance (nullable) is None
        if self.balance is None:
            _dict['balance'] = None

        # set to None if consensus_rewards (nullable) is None
        if self.consensus_rewards is None:
            _dict['consensus_rewards'] = None

        # set to None if execution_rewards (nullable) is None
        if self.execution_rewards is None:
            _dict['execution_rewards'] = None

        # set to None if rewards (nullable) is None
        if self.rewards is None:
            _dict['rewards'] = None

        # set to None if gross_apy (nullable) is None
        if self.gross_apy is None:
            _dict['gross_apy'] = None

        # set to None if updated_at (nullable) is None
        if self.updated_at is None:
            _dict['updated_at'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> EthereumStake:
        """Create an instance of EthereumStake from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return EthereumStake.parse_obj(obj)

        _obj = EthereumStake.parse_obj({
            "validator_address": obj.get("validator_address"),
            "state": obj.get("state"),
            "activated_at": obj.get("activated_at"),
            "activated_epoch": obj.get("activated_epoch"),
            "delegated_block": obj.get("delegated_block"),
            "deposit_tx_sender": obj.get("deposit_tx_sender"),
            "execution_fee_recipient": obj.get("execution_fee_recipient"),
            "withdrawal_credentials": obj.get("withdrawal_credentials"),
            "effective_balance": obj.get("effective_balance"),
            "balance": obj.get("balance"),
            "consensus_rewards": obj.get("consensus_rewards"),
            "execution_rewards": obj.get("execution_rewards"),
            "rewards": obj.get("rewards"),
            "gross_apy": obj.get("gross_apy"),
            "updated_at": obj.get("updated_at")
        })
        return _obj

